#pragma once

#include <AK/HashMap.h>
#include <AK/SinglyLinkedList.h>
#include <Kernel/DoubleBuffer.h>
#include <Kernel/KBuffer.h>
#include <Kernel/Lock.h>
#include <Kernel/Net/IPv4.h>
#include <Kernel/Net/Socket.h>

class IPv4SocketTuple {
public:
    IPv4SocketTuple(IPv4Address local_address, u16 local_port, IPv4Address peer_address, u16 peer_port)
        : m_local_address(local_address)
        , m_local_port(local_port)
        , m_peer_address(peer_address)
        , m_peer_port(peer_port) {};

    IPv4Address local_address() const { return m_local_address; };
    u16 local_port() const { return m_local_port; };
    IPv4Address peer_address() const { return m_peer_address; };
    u16 peer_port() const { return m_peer_port; };

    bool operator==(const IPv4SocketTuple other) const
    {
        return other.local_address() == m_local_address && other.local_port() == m_local_port && other.peer_address() == m_peer_address && other.peer_port() == m_peer_port;
    };

    String to_string() const
    {
        return String::format(
            "%s:%d -> %s:%d",
            m_local_address.to_string().characters(),
            m_local_port,
            m_peer_address.to_string().characters(),
            m_peer_port);
    }

private:
    IPv4Address m_local_address;
    u16 m_local_port { 0 };
    IPv4Address m_peer_address;
    u16 m_peer_port { 0 };
};

namespace AK {

template<>
struct Traits<IPv4SocketTuple> : public GenericTraits<IPv4SocketTuple> {
    static unsigned hash(const IPv4SocketTuple& tuple)
    {
        auto h1 = pair_int_hash(tuple.local_address().to_u32(), tuple.local_port());
        auto h2 = pair_int_hash(tuple.peer_address().to_u32(), tuple.peer_port());
        return pair_int_hash(h1, h2);
    }

    static void dump(const IPv4SocketTuple& tuple)
    {
        kprintf("%s", tuple.to_string().characters());
    }
};

}
